@@ -548,7 +548,8 @@ impl CertD {
     ///
     ///   [The transformation from a fingerprint to a path]: https://www.ietf.org/archive/id/draft-nwjw-openpgp-cert-d-00.html#section-3.2.1
     pub fn get_path_by_fingerprint(&self, fingerprint: &str) -> Result<PathBuf> {
-        if fingerprint.len() != 40 {
+        if ! [pgp::FINGERPRINT_LEN_CHARS_V4,
+              pgp::FINGERPRINT_LEN_CHARS_V6].contains(&fingerprint.len()) {
             return Err(Error::BadName);
         }
         if fingerprint.chars().any(|c| !c.is_ascii_hexdigit()) {
