@@ -40,7 +40,7 @@ impl FileType {
 //
 // For now, we copy Rust's libc's strategy:
 //
-//   https://github.com/rust-lang/rust/blob/96df4943409564a187894018f15f795426dc2e1f/library/std/src/sys/unix/fs.rs#L67
+//   https://github.com/rust-lang/rust/blob/5a2dd7d4f3210629e65879aeecbe643ba3b86bb4/library/std/src/sys/pal/unix/fs.rs#L22
 #[cfg(any(all(target_os = "linux", not(target_env = "musl")), target_os = "hurd"))]
 mod libc64 {
     pub(super) use libc::stat64 as stat;
@@ -132,16 +132,61 @@ impl Metadata {
 /// [`libc::dirent64`](https://docs.rs/libc/latest/libc/struct.dirent64.html)
 pub(crate) struct DirEntry {
     dir: *mut libc::DIR,
+
+    // Get ready for some insanity.
+    //
+    // On some platforms, `dirent.d_name` (or `dirent.d_name`) may be
+    // a fixed-sized array, but the implementation uses a different
+    // structure, which has a flex array.  This means that if we're
+    // not careful about how we use the raw pointer, we may end up
+    // copying the struct, which results in a seg fault when the
+    // underlying struct is smaller than the public struct!  A Rust
+    // libc comment corroborates this:
+    //
+    // > The dirent64 pointers that libc returns from readdir64 are
+    // > allowed to point to allocations smaller _or_ LARGER than
+    // > implied by the definition of the struct.
+    //
+    // https://github.com/rust-lang/rust/blob/5a2dd7d4f3210629e65879aeecbe643ba3b86bb4/library/std/src/sys/pal/unix/fs.rs#L733
+    //
+    // This means that we can't do the following, which copies the
+    // `struct dirent`:
+    //
+    // ```
+    // unsafe { *self.entry }.d_type
+    // ```
+    //
+    // Instead, we need to do the following, which returns a reference
+    // to the `struct dirent`:
+    //
+    // ```
+    // unsafe { (&*self.entry).d_type }
+    // ```
     entry: *mut libc64::dirent,
+
     name_len: OnceCell<usize>,
     // We save the metadata inline to avoid a heap allocation.
     metadata: OnceCell<Result<Metadata>>,
 }
 
 impl DirEntry {
+    // Return `self.entry`'s d_type, safely.
+    //
+    // See the insanity comment above.
+    fn entry_d_type(&self) -> libc::c_uchar {
+        unsafe { (&*self.entry).d_type }
+    }
+
+    // Return `self.entry`'s d_name, safely.
+    //
+    // See the insanity comment above.
+    fn entry_d_name(&self) -> *const c_char {
+        unsafe { (&*self.entry).d_name.as_ptr() as *const c_char }
+    }
+
     /// Returns the file's type, as recorded in the directory.
     pub fn file_type(&self) -> FileType {
-        FileType(unsafe { *self.entry }.d_type)
+        FileType(self.entry_d_type())
     }
 
     /// Returns the filename.
@@ -149,22 +194,10 @@ impl DirEntry {
     /// Note: this is not NUL terminated.
     pub fn file_name(&self) -> &[u8] {
         unsafe {
-            let name = (*self.entry).d_name.as_ptr() as *const c_char;
+            let name = self.entry_d_name();
 
             let name_len = *self.name_len.get_or_init(|| {
-                // According to the Single Unix Specification:
-                //
-                //   The character array d_name is of unspecified
-                //   size, but the number of bytes preceding the
-                //   terminating null byte will not exceed {NAME_MAX}.
-                //
-                // https://pubs.opengroup.org/onlinepubs/007908799/xsh/dirent.h.html
-                //
-                // All platforms that I check use 256 bytes. Don't
-                // hard code that (but do sanity check it).
-                let max_len = std::mem::size_of_val(&(*self.entry).d_name);
-                assert!(max_len >= 128);
-                libc::strnlen(name, max_len)
+                libc::strlen(name)
             });
 
             std::slice::from_raw_parts(
@@ -192,7 +225,7 @@ impl DirEntry {
             let result = unsafe {
                 libc64::fstatat(
                     dirfd,
-                    (*self.entry).d_name.as_ptr() as *const c_char,
+                    self.entry_d_name(),
                     statbuf.as_mut_ptr(),
                     libc::AT_SYMLINK_NOFOLLOW,
                 )
@@ -232,6 +265,7 @@ impl Drop for Dir {
         if let Some(dir) = self.dir.take() {
             unsafe { libc::closedir(dir) };
         }
+        self.entry = None;
     }
 }
 
@@ -267,6 +301,7 @@ impl Dir {
         if entry.is_null() {
             unsafe { libc::closedir(dir) };
             self.dir = None;
+            self.entry = None;
             return None;
         }
 
